/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DebuggerMain.h"
#include "DebugEventServer.h"
#include "Debugger.h"
#include "Configuration.h"
#include "ByteStreamRegion.h"
#include "SymbolManager.h"
#include "DebuggerPluginInterface.h"
#include "DebuggerCoreInterface.h"
#include "DialogAttach.h"
#include "DialogMemoryRegions.h"
#include "Util.h"
#include "DisassemblyLib.h"
#include "i386ArchProcessor.h"
#include "version.h"

#include <QSettings>
#include <QToolButton>
#include <QMessageBox>
#include <QFileDialog>
#include <QInputDialog>
#include <QTableWidgetItem>
#include <QFileInfo>
#include <QTextStream>

#include <fstream>
#include <cassert>

// TODO: move elf parsing functionality to another class
#include <elf.h>

//------------------------------------------------------------------------------
// Name: DebuggerMain(QWidget *parent)
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerMain::DebuggerMain(QWidget *parent) : QMainWindow(parent), 
		m_DebugEventServer(0), m_UpdateGUI(true), m_ReenableBreakPointAddress(0), 
		m_LastEvent(0), m_StepWasRun(false), m_LastOpenDirectory("") {

	ui.setupUi(this);
	
	// setup the global pointers as early as possible
	Debugger::debuggerGUI		= this;
	Debugger::debugEventHandler	= this;

	// make the stack view all nice..
	// TODO: remember what the user set it as, and restore it!
	ui.stackView->setRowWidth(1);
	ui.stackView->setWordWidth(4);
	ui.stackView->setShowAsciiDump(false);
	
	// setup tabs
	setupTabButtons();
	
	// set the listbox to about 4 lines
	QFontMetrics fm(ui.listWidget->font());
	ui.listWidget->setFixedHeight(fm.height() * 5);
	
	// register our DebugEvent type so it can be used in 
	// inter-thread signals
	qRegisterMetaType<DebugEvent>("DebugEvent");
	
	// make sure our widgets use custom context menus
	ui.registerList->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.cpuView->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.stackView->setContextMenuPolicy(Qt::CustomContextMenu);
			
	// add toggles for the dock windows
	ui.menu_View->addAction(ui.registersDock->toggleViewAction());
	ui.menu_View->addAction(ui.dataDock->toggleViewAction());
	ui.menu_View->addAction(ui.stackDock->toggleViewAction());

	// remove the one in the designer and put in our built one
	// it's a bit ugly to do it this way, but the designer wont let me 
	// make a tabless entry..and the ones i create look slightly diff 
	// (probably lack of layout manager in mine...
	ui.tabWidget->removeTab(0);
	createNewDataTab();
	
	m_StackViewInfo.view = ui.stackView;
	
	QFont stackFont;
	stackFont.fromString(Debugger::config()->StackViewFont);
	ui.stackView->setFont(stackFont);
	
	m_ArchProcessor = new i386ArchProcessor;
	m_ArchProcessor->setupRegisterView(ui.registerList);
	m_ArchProcessor->updateRegisterView("", m_CurrentState);
	
	m_WorkingDirectory = QDir().absolutePath();
}


//------------------------------------------------------------------------------
// Name: ~DebuggerMain()
// Desc: destructor
//------------------------------------------------------------------------------
DebuggerMain::~DebuggerMain() {

	switch(Debugger::config()->closeBehaviour) {
	case Configuration::Terminate:
		on_action_Kill_triggered();
		break;
	case Configuration::Detach:
		on_action_Detach_triggered();
		break;
	}
	
	delete m_ArchProcessor;
	
	// remove the last tab
	removeDataTab();
	
	// clean up the stream for the stack view too
	safeDelete(m_StackViewInfo.stream);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::loadRecentFileList() {
	QSettings settings;
	settings.beginGroup("Recent Files");
	
	for(int i = 0; i < MaxRecentFiles; ++i) {
		const QString s = settings.value(QString().sprintf("File%d", i), "").toString();
		if(!s.isEmpty()) {
			m_RecentFileList.push_back(s);
		}
	}
	settings.endGroup();
	updateRecentFileList();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::saveRecentFileList() {
	QSettings settings;
	settings.beginGroup("Recent Files");
	
	for(int i = 0; i < MaxRecentFiles; ++i) {
		QString s;
		if(i < m_RecentFileList.size()) {
			s = m_RecentFileList[i];
		}
		settings.setValue(QString().sprintf("File%d", i), s);
	}
	settings.endGroup();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::openRecentMenuItem() {
	QAction *const action = qobject_cast<QAction *>(sender());
	if(action != 0) {
		const QString s = action->data().toString();
		doOpenFile(s);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateRecentFileList() {

	QMenu *menu = new QMenu(this);
	for(QVector<QString>::iterator it = m_RecentFileList.begin(); it != m_RecentFileList.end(); ++it) {
		QAction *action = menu->addAction(*it, this, SLOT(openRecentMenuItem()));
		if(action != 0) {
			action->setData(*it);
		}
	}
	menu->addSeparator();
	menu->addAction("Clear &Menu", this, SLOT(clearRecentMenu()));
	ui.action_Recent_Files->setMenu(menu);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::clearRecentMenu() {
	m_RecentFileList.clear();
	updateRecentFileList();
	
}

//------------------------------------------------------------------------------
// Name: finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins)
// Desc: finalizes plugin setup by adding each to the menu, we can do this now 
//       that we have a GUI widget to attach it to
//------------------------------------------------------------------------------
void DebuggerMain::finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins) {
	// setup the menu for all plugins that which to do so
	foreach (DebuggerPluginInterface *plugin, plugins) {
		QMenu *const menu = plugin->menu(this);
		if(menu != 0) {
			ui.menu_Plugins->addMenu(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: closeEvent(QCloseEvent *)
// Desc: triggered on main window close, saves window state
//------------------------------------------------------------------------------
void DebuggerMain::closeEvent(QCloseEvent *) {
	saveWindowState();
	saveRecentFileList();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::loadWindowState() {
	QSettings settings;
	settings.beginGroup("Window");
	const QByteArray state = settings.value("window_state").toByteArray();
	
	const int width = settings.value("window_width", -1).toInt();
	const int height = settings.value("window_height", -1).toInt();
	if(width != -1) {
		resize(width, size().height());
	}
	
	if(height != -1) {
		resize(size().width(), height);
	}
	
	settings.endGroup();
	restoreState(state);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::saveWindowState() {
	QSettings settings;
	const QByteArray state = saveState();
	settings.beginGroup("Window");
	settings.setValue("window_state", state);
	settings.setValue("window_width", width());
	settings.setValue("window_height", height());
	settings.endGroup();
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: triggered on show, restores window state
//------------------------------------------------------------------------------
void DebuggerMain::showEvent(QShowEvent *) {
	loadWindowState();
	loadRecentFileList();
}

//------------------------------------------------------------------------------
// Name: setupTabButtons()
// Desc: creates the add/remove tab buttons in the data view
//------------------------------------------------------------------------------
void DebuggerMain::setupTabButtons() {
	// add the corner widgets to the data view
	m_AddTab = new QToolButton(ui.tabWidget);
	m_DelTab = new QToolButton(ui.tabWidget);
	
	m_AddTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	m_DelTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	
	m_AddTab->setIcon(QIcon(":/debugger/pics/addtab.png"));
	m_DelTab->setIcon(QIcon(":/debugger/pics/closetab.png"));
	m_AddTab->setAutoRaise(true);
	m_DelTab->setAutoRaise(true);
	m_DelTab->setEnabled(false);
	
	ui.tabWidget->setCornerWidget(m_AddTab, Qt::TopLeftCorner);
	ui.tabWidget->setCornerWidget(m_DelTab, Qt::TopRightCorner);
	
	connect(m_AddTab, SIGNAL(clicked()), this, SLOT(createNewDataTab()));
	connect(m_DelTab, SIGNAL(clicked()), this, SLOT(removeDataTab()));
}

//------------------------------------------------------------------------------
// Name: on_registerList_customContextMenuRequested(const QPoint &pos)
// Desc: context menu handler for register view
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_customContextMenuRequested(const QPoint &pos) {
	QTreeWidgetItem *const item =  ui.registerList->itemAt(pos);
	if(item != 0) {
		if(!ui.registerList->isCatagory(item)) {
			QMenu *menu = new QMenu(this);
			menu->addAction("&Follow In Dump", this, SLOT(followInDump()));
			menu->addAction("&Follow In Dump (New Tab)", this, SLOT(followInDumpNewTab()));
			menu->addAction("&Follow In Stack", this, SLOT(followInStack()));
			menu->exec(ui.registerList->mapToGlobal(pos));
			safeDelete(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_breakPointToggled(BaseTypes::address_t address)
// Desc: handler for toggling the breakpoints
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_breakPointToggled(BaseTypes::address_t address) {
	Debugger::toggleBreakPoint(address);
}

//------------------------------------------------------------------------------
// Name: valueFromName(const QString &s, unsigned int mode)
// Desc: gets the value of a register based on name, takes into account a "mode"
//       mask, which lets us select catagories of registers to consider valid
// Note: portability issue, implementation uses x86 specific names
//------------------------------------------------------------------------------
BaseTypes::reg_t *DebuggerMain::valueFromName(const QString &s, unsigned int mode) {
	
	if(mode & State::EFLAGS_FLAG && s.startsWith("EFLAGS:"))	return &m_CurrentState.eflags;
	if(mode & State::EAX_FLAG && s.startsWith("EAX:")) 			return &m_CurrentState.eax;
	if(mode & State::EBX_FLAG && s.startsWith("EBX:")) 			return &m_CurrentState.ebx;
	if(mode & State::ECX_FLAG && s.startsWith("ECX:")) 			return &m_CurrentState.ecx;
	if(mode & State::EDX_FLAG && s.startsWith("EDX:")) 			return &m_CurrentState.edx;
	if(mode & State::EBP_FLAG && s.startsWith("EBP:")) 			return &m_CurrentState.ebp;
	if(mode & State::ESP_FLAG && s.startsWith("ESP:")) 			return &m_CurrentState.esp;
	if(mode & State::EDI_FLAG && s.startsWith("EDI:")) 			return &m_CurrentState.edi;;
	if(mode & State::ESI_FLAG && s.startsWith("ESI:")) 			return &m_CurrentState.esi;
	if(mode & State::EIP_FLAG && s.startsWith("EIP:")) 			return &m_CurrentState.eip;
	if(mode & State::CS_FLAG && s.startsWith("CS:")) 			return &m_CurrentState._cs;
	if(mode & State::DS_FLAG && s.startsWith("DS:")) 			return &m_CurrentState._ds;
	if(mode & State::ES_FLAG && s.startsWith("ES:")) 			return &m_CurrentState._es;
	if(mode & State::FS_FLAG && s.startsWith("FS:")) 			return &m_CurrentState._fs;
	if(mode & State::GS_FLAG && s.startsWith("GS:")) 			return &m_CurrentState._gs;
	if(mode & State::SS_FLAG && s.startsWith("SS:")) 			return &m_CurrentState._ss;
	
	return 0;
}

//------------------------------------------------------------------------------
// Name: followInStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followInStack() {
	const QTreeWidgetItem *const i = ui.registerList->currentItem();
	if(i != 0) {
		const QString s = i->text(0);
		const unsigned int *value = valueFromName(s, State::GPR_FLAGS);

		if(value != 0) {
			if(!Debugger::dumpStack(*value)) {
				QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(*value)));
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: followRegisterInDump(bool tabbed)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followRegisterInDump(bool tabbed) {
	const QTreeWidgetItem *const i = ui.registerList->currentItem();
	if(i != 0) {
		const QString s = i->text(0);
		const unsigned int *value = valueFromName(s, State::GPR_FLAGS);

		if(value != 0) {
			if(!Debugger::dumpData(*value, tabbed)) {
				QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(*value)));
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_registerList_itemDoubleClicked(QTreeWidgetItem *item)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_itemDoubleClicked(QTreeWidgetItem *item) {

	const QString s = item->text(0);
	unsigned int *value = valueFromName(s, State::ANY_FLAGS);
	
	if(value != 0) {
		if(Debugger::getValueFromUser(*value, "Register Value")) {			
			Debugger::debuggerBase->setState(m_CurrentState);
			updateGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_About_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_About_triggered() {

	QMessageBox::about(this, "About EDB", QString().sprintf(
		"<p>EDB (Evan's Debugger) is designed to be an easy to use, modular, and cross platform debugger.</p>"
		"<p>More information and updates can be found at <a href=\"http://www.codef00.com\">http://www.codef00.com</a></p>"
		"<p>Written by Evan Teran.</p>"
		"<p>version: %s</p>", version_string));
}

//------------------------------------------------------------------------------
// Name: on_actionAbout_QT_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionAbout_QT_triggered() {
	QMessageBox::aboutQt(this, "About QT");
}

//------------------------------------------------------------------------------
// Name: doAddressJump(BaseTypes::address_t address)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerMain::doAddressJump(BaseTypes::address_t address) {
	bool ret = Debugger::jumpToAddress(address);
	if(!ret) {
		QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToAddress() {
    bool ok;
    QString text = QInputDialog::getText(this, "Goto Address", "Address:", QLineEdit::Normal, "", &ok);
    if (ok && !text.isEmpty()) {
		const BaseTypes::address_t address = text.toUInt(&ok, 16);
		if(ok) {
			doAddressJump(address);
		}
	}
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToEIP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToEIP() {
	doAddressJump(m_CurrentState.eip);
}

//------------------------------------------------------------------------------
// Name: on_action_Configure_Debugger_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Configure_Debugger_triggered() {
	
	if(Debugger::dialogOptions()->exec() == QDialog::Accepted) {	
		// reload symbols incase they changed

		safeDelete(Debugger::symbolManager);
		Debugger::symbolManager = new SymbolManager(Debugger::config()->symbolPath);

		if(Debugger::memoryRegions != 0) {
			Debugger::memoryRegions->sync();
		}
		
		QFont stackFont;
		stackFont.fromString(Debugger::config()->StackViewFont);
		ui.stackView->setFont(stackFont);

		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_customContextMenuRequested(const QPoint &pos) {
	QMenu *menu = new QMenu(this);
	// TODO: add comment
	//menu->addAction("Add &Comment", this, SLOT(""));
	//menu->addSeparator();
	menu->addAction("&Goto Address", this, SLOT(mnuCPUJumpToAddress()));
	menu->addAction("&Goto EIP", this, SLOT(mnuCPUJumpToEIP()));
	
	menu->exec(ui.cpuView->viewport()->mapToGlobal(pos));
	safeDelete(menu);
}

//------------------------------------------------------------------------------
// Name: mnuGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuGotoAddress() {

    bool ok;
    QString text = QInputDialog::getText(this, "Goto Address", "Address:", QLineEdit::Normal, "", &ok);
    if (ok && !text.isEmpty()) {
		const BaseTypes::address_t address = text.toUInt(&ok, 16);
		if(ok) {		
			if(!Debugger::dumpData(address)) {
				QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
			}
		}
	}
}


//------------------------------------------------------------------------------
// Name: doFollowInDump(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInDump(const QHexView *hv) {
	if(hv != 0) {
		if(hv->hasSelectedText()) {
			const QByteArray data = hv->selectedBytes();

			if(data.size() == sizeof(BaseTypes::address_t)) {
				union {
					quint8	b[sizeof(BaseTypes::address_t)];
					quint32	d;
				} dword;

				for(unsigned int i = 0; i < sizeof(BaseTypes::address_t); ++i) {
					dword.b[i] = data[i];
				}

				if(!Debugger::dumpData(dword.d)) {
					QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(dword.d)));
				}
			} else {
				QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(BaseTypes::address_t)));
			}
		} else {
			QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(BaseTypes::address_t)));
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuFollowInDump() {
	doFollowInDump(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInDump() {
	doFollowInDump(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Arguments_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Arguments", "Arguments:", QLineEdit::Normal, m_Arguments, &ok);
	if(ok) {
		m_Arguments = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Working_Directory_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Working Directory", "Working Directory:", QLineEdit::Normal, m_WorkingDirectory, &ok);
	if(ok) {
		m_WorkingDirectory = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::pushDWORD() {

	uint32_t value;
	
	Debugger::debuggerBase->readBytes(m_CurrentState.esp - sizeof(value), &value, sizeof(value));

	if(Debugger::getValueFromUser(value, "Enter value to push")) {

		m_CurrentState.esp -= sizeof(value);
		Debugger::debuggerBase->writeBytes(m_CurrentState.esp, &value, sizeof(value));

		Debugger::debuggerBase->setState(m_CurrentState);
		updateGUI();
	}
	
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
// Note: portability issue, assumes data is 4 bytes wide
//------------------------------------------------------------------------------
void DebuggerMain::popDWORD() {
	m_CurrentState.esp += 4;
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_stackView_customContextMenuRequested(const QPoint &pos) {
	QHexView *const s = ui.stackView;
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuStackFollowInDump()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(modifyBytesStack()));
		menu->addSeparator();
		menu->addAction("&Push DWORD", this, SLOT(pushDWORD()));
		menu->addAction("P&op DWORD", this, SLOT(popDWORD()));
		menu->exec(s->mapToGlobal(pos));
		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::dumpDataToFile() {
	QHexView *const s = qobject_cast<QHexView *>(ui.tabWidget->currentWidget());
	if(s != 0) {
		
		const QString filename = QFileDialog::getSaveFileName(this, "Save File", m_LastOpenDirectory, "");
		if(!filename.isEmpty()) {
			
			QFile file(filename);
			file.open(QIODevice::WriteOnly);
			if(file.isOpen()) {
				file.write(s->allBytes());
			}			
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytes(QHexView *s) {
	if(s != 0) {
		const BaseTypes::address_t address = s->selectedBytesAddress();
		const unsigned int size = s->selectedBytesSize();
		QByteArray bytes = s->selectedBytes();
		
		
		if(Debugger::getBinaryStringFromUser(bytes, "Edit Binary String", size)) {
		
			while(bytes.size() < static_cast<int>(size)) {
				bytes.push_back('\x00');
			}
		
			Debugger::debuggerBase->writeBytes(address, bytes.data(), size);
			updateGUI();
		}	
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::modifyBytesDump() {
	doModifyBytes(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::modifyBytesStack() {
	doModifyBytes(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::dumpContextMenu(const QPoint &pos) {
	QHexView *const s = qobject_cast<QHexView *>(sender());
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuFollowInDump()));
		menu->addAction("&Goto Address", this, SLOT(mnuGotoAddress()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(modifyBytesDump()));
		menu->addSeparator();
		menu->addAction("&Dump To File", this, SLOT(dumpDataToFile()));
		menu->exec(s->mapToGlobal(pos));

		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateMenuState(GUI_STATE guiState) {

	switch(guiState) {
	case PAUSED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(true);
	#endif
		ui.action_Restart->setEnabled(true);
		ui.action_Run->setEnabled(true);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(true);
		ui.action_Step_Over->setEnabled(true);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(true);
		break;
	case RUNNING:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(true);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		break;
	case TERMINATED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		break;
	}
	m_GUIState = guiState;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventSignaled() {
	switch(m_LastEvent.exitSignal()) {
	case DebugEvent::sigkill:
		QMessageBox::information(this, "Application Killed", "The debugged application was killed.");
		break;
	default:
		Debugger::log(QString().sprintf("Unknown Termination Signal Recieved! %d", m_LastEvent.exitSignal()));
		break;
	}

	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventStopped() {


	// ok we just came in from a signal, we need to test some things, 
	// generally, we will want to check if it was a step, if it was, was it because
	// we just hit a break point or because we wanted to run, but had to step first
	// in case were were on a breakpoint already...

	switch(m_LastEvent.stopSignal()) {
	case DebugEvent::sigsegv:
		QMessageBox::information(this, "Illegal Access Fault", QString().sprintf(
			"<p>The debugged application encountered a segmentation fault!<br />"
			"The address <strong>%p</strong> could not be accessed.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>", 
			reinterpret_cast<void *>(Debugger::debuggerBase->getFaultAddress()))
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigill:
		QMessageBox::information(this, "Illegal Instruction Fault", 
			"<p>The debugged application attempted to execute an illegal instruction.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigfpe:
		QMessageBox::information(this, "Floating Point Exception", 
			"<p>The debugged application encountered a floating-point exception!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigstop:
		// user asked to pause the debugged process
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigtrap:
		
		// we just got a SIGTRAP, there are a few possible causes
		// #1: we just did an execve, in which case we want to stop cause we just attached
		// #2: we hit a 0xCC breakpoint, if so, then we want to stop
		// #3: we did a step
		
		do {
			// prevEIP will have the address of the instruction which brought us here!
			const BaseTypes::address_t prevEIP = m_CurrentState.eip - Debugger::debuggerBase->executionBreakPointSize();
			
			// look it up in our breakpoint list, make sure it is one of OUR int3s!
			// if it is, we need to backup EIP and pause ourselves
			const BreakPoint *const bp = Debugger::debuggerBase->findBreakPoint(prevEIP);
			if(bp != 0 && bp->enabled()) {
						
				// if it's a one time breakpoint then we should remove it upon
				// triggering, this is mainly used for "run-until" situations
				// like step over

				if(bp->oneTime()) {
					Debugger::debuggerBase->removeBreakpoint(prevEIP);
				} else {
					// TODO: make this work
					//++bp->passcount;					
				}

				// back up eip, since we executed a breakpoint instead of 
				// the real code that belongs there
				m_CurrentState.eip = prevEIP;
				Debugger::debuggerBase->setState(m_CurrentState);

				m_StepWasRun = false;
			} else {
				// it wasn't one of our int3s, so what do we do?

				// this code attempts to cancel a run which will
				// cause overrun a breakpoint potentially				
				uint8_t breakpointData[Debugger::debuggerBase->executionBreakPointSize()];
				if(Debugger::debuggerBase->readBytes(prevEIP, breakpointData, sizeof(breakpointData))) {
					if(std::memcmp(
						breakpointData, 
						Debugger::debuggerBase->executionBreakPointData().data(), 
						Debugger::debuggerBase->executionBreakPointSize()) == 0) {
						m_StepWasRun = false;			
					}
				}
			}
		} while(0);
		break;

	case DebugEvent::sigchld:
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
		
	case DebugEvent::sigcont:
		// this one is probably not very interesting, so we'll just resume
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		// we do an early return so it's like it never happened
		return;
		
	default:
		Debugger::log(QString().sprintf("Unknown Stop Signal Recieved! %d", m_LastEvent.stopSignal()));
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
	}

	// re-enable any breakpoints we disabled
	if(m_ReenableBreakPointAddress != 0) {
		Debugger::debuggerBase->enableBreakpoint(m_ReenableBreakPointAddress);
		m_ReenableBreakPointAddress = 0;
	}

	// if we stepped with the intention of running, well we should run then!
	if(m_StepWasRun) {
		m_StepWasRun = false;
		Debugger::debuggerBase->resume(0);
	} else {
		updateGUI();
		updateMenuState(PAUSED);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventExited() {
	QMessageBox::information(this, "Application Exited", QString().sprintf("The debugged application exited normally with exit code %d.",  m_LastEvent.exitStatus()));
	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEvent(DebugEvent event) {
	if(Debugger::debuggerBase != 0) {
		Debugger::debuggerBase->getState(m_CurrentState);
	}
	
	// it is possible but rare that that we get an event when no memory regions
	// are setup (on detach)
	if(Debugger::memoryRegions != 0) {
		Debugger::memoryRegions->sync();
	}
	
	m_LastEvent = event;
	
	switch(m_LastEvent.reason()) {
	case DebugEvent::EVENT_STOPPED:
		handleEventStopped();
		break;
		
	case DebugEvent::EVENT_SIGNALED:
		handleEventSignaled();
		break;
		
	case DebugEvent::EVENT_EXITED:
		handleEventExited();
		break;
		
	case DebugEvent::EVENT_UNKNOWN:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::debugEventHandler(DebugEvent event) {

	if(Debugger::debugEventHandler != 0) {
		Debugger::debugEventHandler->handleEvent(event);
	} else {
		assert(0);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::DataViewInfo::update() {

	safeDelete(stream);
	stream = new ByteStreamRegion(region);
	
	if(view != 0) {
		view->setAddressOffset(region.start);
		view->setData(stream);
	}
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateData(DebuggerMain::DataViewInfo *v) {
	if(v != 0) {
		v->update();	
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", v->region.start, v->region.end));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::clearData(DataViewInfo *v) {
	if(v != 0) {
		v->view->clear();
		v->view->scrollTo(0);
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", v->region.start, v->region.end));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDisassembly(BaseTypes::address_t address, const MemoryRegions::Region &r) {

	ui.cpuView->setCurrentAddress(m_CurrentState.eip);
	ui.cpuView->setAddressOffset(r.start);
	ui.cpuView->setRegion(r);
	
	if(!ui.cpuView->addressShown(m_CurrentState.eip)) {
		ui.cpuView->scrollTo(address);
	}
	
	ui.listWidget->clear();
	ui.listWidget->addItems(m_ArchProcessor->updateInstructionInfo(address, m_CurrentState));
	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateStackView() {
	if(Debugger::memoryRegions == 0 || !Debugger::dumpStack(m_CurrentState.esp)) {
		ui.stackView->clear();
		ui.stackView->scrollTo(0);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateCPUView(MemoryRegions::Region &region) {
	if(Debugger::memoryRegions == 0 || !Debugger::memoryRegions->findRegion(m_CurrentState.eip, region)) {
		ui.cpuView->clear();
		ui.cpuView->scrollTo(0);
		ui.listWidget->clear();
	} else {
		updateDisassembly(m_CurrentState.eip, region);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDataViews() {
	if(Debugger::memoryRegions != 0) {
		// update all data views with the current region data
		foreach(DataViewInfo *info, m_DataRegions) {

			// make sure the regions are still valid..
			if(Debugger::memoryRegions->findRegion(info->region.start)) {
				updateData(info);
			} else {
				clearData(info);
			}
		}
	} else {
		// clear all data views
		foreach(DataViewInfo *info, m_DataRegions) {
			clearData(info);
		}
	}
}

//------------------------------------------------------------------------------
// Name: updateGUI()
// Desc: updates all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::updateGUI() {

	if(m_UpdateGUI) {
	
		MemoryRegions::Region region;

		updateCPUView(region);
		updateDataViews();
		updateStackView();
		m_ArchProcessor->updateRegisterView(region.name, m_CurrentState);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doResume(bool passExceptionToApp) {
	
	// are we currently sitting on a breakpoint
	const bool atBreakPoint = Debugger::debuggerBase->findBreakPoint(m_CurrentState.eip) != 0;
	
	// if we are on a breakpoint, disable it, and set the reenable 
	// address to the current eip
	if(atBreakPoint) {
		Debugger::debuggerBase->disableBreakpoint(m_CurrentState.eip);
		m_ReenableBreakPointAddress = m_CurrentState.eip;
	}

	// set the state to "running"
	updateMenuState(RUNNING);
	
	// if neccessary pass the signal to the application, otherwise just step
	if(passExceptionToApp && m_LastEvent.stopped() && m_LastEvent.stopSignal() != SIGTRAP && m_LastEvent.stopSignal() != SIGSTOP) {
		Debugger::debuggerBase->step(m_LastEvent.stopSignal());
	} else {
		Debugger::debuggerBase->step(0);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered() {
	m_StepWasRun = true;
	doResume(true);	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered() {
	doResume(true);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_Pass_Signal_To_Application_triggered() {
	uint8_t buffer[MAX_INSN_SIZE];
	
	Instruction insn;
	Debugger::debuggerBase->readBytes(m_CurrentState.eip, buffer, sizeof(buffer));
		
	const int size = insn.disassemble(buffer, sizeof(buffer), 0, 0);
	
	if(size != 0 && (insn.type() == insn_call || insn.type() == insn_callcc || insn.prefix() & (insn_rep_zero | insn_rep_notzero))) {
		// add a temporary breakpoint at the instruction just after the call
		Debugger::debuggerBase->addBreakpoint(m_CurrentState.eip + size, true);
		on_action_Run_Pass_Signal_To_Application_triggered();
	} else {
		on_action_Step_Into_Pass_Signal_To_Application_triggered();
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_triggered() {
	m_StepWasRun = true;
	doResume(false);	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_triggered() {
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_triggered() {

	uint8_t buffer[MAX_INSN_SIZE];
	Instruction insn;

	Debugger::debuggerBase->readBytes(m_CurrentState.eip, buffer, sizeof(buffer));
		
	const int size = insn.disassemble(buffer, sizeof(buffer), 0, 0);
	
	if(size != 0 && (insn.type() == insn_call || insn.type() == insn_callcc || insn.prefix() & (insn_rep_zero | insn_rep_notzero))) {
		// add a temporary breakpoint at the instruction just after the call
		Debugger::debuggerBase->addBreakpoint(m_CurrentState.eip + size, true);
		on_action_Run_triggered();
	} else {
		on_action_Step_Into_triggered();
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionRun_Until_Return_triggered() {
	// TODO: implement this
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Pause_triggered() {
	Debugger::debuggerBase->pause();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Restart_triggered() {

	if(Debugger::debuggerBase != 0) {
	
		m_WorkingDirectory = getProcessCurrentWorkingDirectory(Debugger::debuggerBase->pid());
	
		QStringList args = getProcessArguments(Debugger::debuggerBase->pid());
		
		if(args.size() != 0) {
		
			const QString s = args[0];
			args.removeFirst();

			if(!s.isEmpty()) {
				doDetachFromProcess(true);

				if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args)) {
					setInitialDebuggerState();
					setInitialBreakpoint(s);
					setWindowTitle(QString().sprintf("EDB - %s", qPrintable(s)));
				} else {
					QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
				}

				Debugger::debuggerBase->getState(m_CurrentState);
				updateGUI();
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupEventThread() {

	if(m_DebugEventServer != 0) {
		
		// make sure we don't get any spurious events!
		disconnect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
		
		// make sure it thinks it shoudl exit
		m_DebugEventServer->m_ShouldExit = true;

		// this is ugly, but it seems to work pretty well
		// basically we have marked the thread as "shoudl exit"
		// but, we may be blocked (likley) waiting for the next event
		// so we intentionally cause one last event giving it a change to
		// check the loop invariant one more time, which will allow
		// the thread to exit cleanly
		if(m_GUIState == RUNNING) {
			Debugger::debuggerBase->pause();
			updateMenuState(PAUSED);
		} else {
			Debugger::debuggerBase->step(0);
			updateMenuState(RUNNING);
		}
		m_DebugEventServer->wait();

		// TODO: find a way to wake up the waiting event thread
		m_DebugEventServer = 0;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupDebugger() {
	
	cleanupEventThread();

	safeDelete(Debugger::memoryRegions);
	safeDelete(Debugger::symbolManager);

	// clear up the data view
	while(ui.tabWidget->count() > 1) {
		removeDataTab();
	}
	
	setWindowTitle("EDB");
	
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doDetachFromProcess(bool kill) {

	cleanupDebugger();
	
	m_StepWasRun = false;
	
	if(Debugger::debuggerBase != 0) {
		if(kill) {
			Debugger::debuggerBase->kill();
		} else {
			Debugger::debuggerBase->detach();
		}
	}
	
	std::memset(&m_CurrentState, 0, sizeof(m_CurrentState));
	
	updateMenuState(TERMINATED);
}

//------------------------------------------------------------------------------
// Name: on_action_Detach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Detach_triggered() {
	doDetachFromProcess(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Kill_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Kill_triggered() {
	doDetachFromProcess(true);
}

//------------------------------------------------------------------------------
// Name: getELFEntryPoint(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
BaseTypes::address_t DebuggerMain::getELFEntryPoint(const QString &s) {

	union ElfHeader {
		Elf32_Ehdr elf32;
		Elf64_Ehdr elf64;
	};

	std::ifstream file(qPrintable(s), std::ios::binary);
	if(file) {
		ElfHeader header;
		file.seekg(0);
		if(file.read(reinterpret_cast<char *>(&header), sizeof(header))) {
			if(std::memcmp(header.elf32.e_ident, ELFMAG, SELFMAG) == 0) {
				switch(header.elf32.e_ident[EI_CLASS]) {
				case ELFCLASS32:
					return header.elf32.e_entry;
					break;
				case ELFCLASS64:
					// NOT YET SUPPORTED
					assert(0);
					return header.elf64.e_entry;
					break;
				}
			}
		}
	}
	return 0;
}



//------------------------------------------------------------------------------
// Name: setInitialDebuggerState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialDebuggerState() {

	updateMenuState(PAUSED);
		
	m_DebugEventServer = new DebugEventServer(this);
	connect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
	m_DebugEventServer->start();
	
	Debugger::symbolManager = new SymbolManager(Debugger::config()->symbolPath);
	Debugger::memoryRegions = new MemoryRegions(Debugger::debuggerBase->pid());
	Debugger::memoryRegions->sync();
	
	const QVector<MemoryRegions::Region> &l = Debugger::memoryRegions->regions();

	assert(m_DataRegions.size() > 0);

	if(l.size() >= 2) {
		m_DataRegions[0]->region = l[1];
	} else {
		m_DataRegions[0]->region = MemoryRegions::Region();
	}
	
	std::memset(&m_CurrentState, 0, sizeof(m_CurrentState));
	
	m_StepWasRun = false;
}

//------------------------------------------------------------------------------
// Name: setInitialBreakpoint(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialBreakpoint(const QString &s) {
	BaseTypes::address_t entryPoint = 0;
	if(Debugger::config()->initialBreakPoint == Configuration::MainSymbol) {

		const QFileInfo fileInfo(s);
		QString mainSymbol = fileInfo.baseName();
		if(!fileInfo.completeSuffix().isEmpty()) {
			mainSymbol += ".";
			mainSymbol += fileInfo.completeSuffix();
		}
		
		mainSymbol += "::main";

		const SymbolManager::Symbol *const sym = Debugger::symbolManager->findSymbol(mainSymbol);

		if(sym != 0) {
			entryPoint = sym->address;
		} else {
			entryPoint = getELFEntryPoint(s);
		}
	} else if(Debugger::config()->initialBreakPoint == Configuration::EntryPoint) {
		entryPoint = getELFEntryPoint(s);
	}

	if(entryPoint != 0) {
		Debugger::debuggerBase->addBreakpoint(entryPoint);
	}
}

//------------------------------------------------------------------------------
// Name: doOpenFile(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doOpenFile(const QString &s) {
	if(!s.isEmpty()) {
		m_LastOpenDirectory = QFileInfo(s).canonicalFilePath();
	
		on_action_Detach_triggered();

		QStringList args = m_Arguments.split(" ", QString::SkipEmptyParts);
	
		if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args)) {
			setInitialDebuggerState();
			setInitialBreakpoint(s);
			setWindowTitle(QString().sprintf("EDB - %s", qPrintable(s)));
			
			// update recent file list
			if(!m_RecentFileList.contains(s)) {
				m_RecentFileList.push_front(s);
				
				// make sure we don't add more than the max
				while(m_RecentFileList.size() > MaxRecentFiles) {
					m_RecentFileList.pop_back();
				}
				
				updateRecentFileList();
			}	
		} else {
			QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
		}
		
		Debugger::debuggerBase->getState(m_CurrentState);
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Open_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Open_triggered() {
	doOpenFile(QFileDialog::getOpenFileName(this, "Choose a file", m_LastOpenDirectory, ""));
}

//------------------------------------------------------------------------------
// Name: doAttach(BaseTypes::pid_t pid)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doAttach(BaseTypes::pid_t pid) {
	if(pid != Debugger::debuggerBase->pid()) {

		on_action_Detach_triggered();

		if(Debugger::debuggerBase->attach(pid)) {
		
			m_WorkingDirectory = getProcessCurrentWorkingDirectory(pid);
		
			QStringList args = getProcessArguments(pid);

			setInitialDebuggerState();
			setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(args[0]), pid));
			args.removeFirst();
			
			// make our arguments match what it was run with
			m_Arguments = args.join(" ");
		} else {
			QMessageBox::information(this, "EDB", "Failed to attach to process, please check priviledges and try again.");				
		}

		Debugger::debuggerBase->getState(m_CurrentState);
		updateGUI();
	} else {
		QMessageBox::information(this, "EDB", "You are already debugging that process!");
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Attach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Attach_triggered() {
	static DialogAttach *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogAttach(this);
	} 
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const BaseTypes::pid_t pid = sel.begin()[0]->text().toUInt();
			doAttach(pid);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Memory_Regions_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Memory_Regions_triggered() {
	static DialogMemoryRegions *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogMemoryRegions(this);
	}
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const BaseTypes::address_t region_start	= sel.begin()[0]->text().toUInt(0, 16);
			Debugger::dumpData(region_start);
		}
	}
}

//------------------------------------------------------------------------------
// Name: createNewDataTab()
// Desc: duplicates the current tab creating a new one
//------------------------------------------------------------------------------
void DebuggerMain::createNewDataTab() {
	
	const int currentIndex = ui.tabWidget->currentIndex();
	
	DataViewInfo *newDataView = new DataViewInfo;
	
	if(currentIndex != -1) {
		// duplicate the current region
		newDataView->region = m_DataRegions[currentIndex]->region;
	}
	
	newDataView->stream	= new ByteStreamRegion(newDataView->region);
	newDataView->view	= new QHexView;
		
	// setup the context menu
	newDataView->view->setContextMenuPolicy(Qt::CustomContextMenu);
	connect(newDataView->view, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(dumpContextMenu(const QPoint &)));
	
	newDataView->view->setAddressOffset(newDataView->region.start);
	newDataView->view->setData(newDataView->stream);
	
	m_DataRegions.push_back(newDataView);

	// create the tab!
	ui.tabWidget->addTab(newDataView->view, QString().sprintf("%08x-%08x", newDataView->region.start, newDataView->region.end));
	ui.tabWidget->setCurrentIndex(ui.tabWidget->count() - 1);
	
	if(ui.tabWidget->count() > 1) {
		m_DelTab->setEnabled(true);
	}
}

//------------------------------------------------------------------------------
// Name: removeDataTab() 
// Desc: handles removing of a memory view tab
//------------------------------------------------------------------------------
void DebuggerMain::removeDataTab() {
	const int currentIndex = ui.tabWidget->currentIndex();
	
	DataViewInfo *info = m_DataRegions[currentIndex];
	m_DataRegions.remove(currentIndex);
	
	ui.tabWidget->removeTab(currentIndex);
	if(ui.tabWidget->count() <= 1) {
		m_DelTab->setEnabled(false);
	}

	if(info != 0) {
		safeDelete(info->stream);
		safeDelete(info->view);
	}
	
	safeDelete(info);
}

//------------------------------------------------------------------------------
// Name: getProcessArguments(BaseTypes::pid_t pid)
// Desc: 
// Note: Linux specific
//------------------------------------------------------------------------------
QStringList DebuggerMain::getProcessArguments(BaseTypes::pid_t pid) {
	const QString cmdlineFile(QString().sprintf("/proc/%d/cmdline", pid));
	QFile file(cmdlineFile);
	QStringList ret;
	if(file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		QTextStream in(&file);
		
		QString s;
		QChar ch;
		
		while(in.status() == QTextStream::Ok) {
			in >> ch;
			if(ch == '\0') {
				if(!s.isEmpty()) {
					ret.append(s);
				}
				s.clear();
			} else {
				s += ch;
			}
		}
		
		if(!s.isEmpty()) {
			ret.append(s);
		}
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: getProcessCurrentWorkingDirectory(BaseTypes::pid_t pid)
// Desc: 
// Note: Linux specific
//------------------------------------------------------------------------------
QString DebuggerMain::getProcessCurrentWorkingDirectory(BaseTypes::pid_t pid) {
	
	char buf[PATH_MAX];
	QString ret;
	ssize_t i;
	const QString cwdFile(QString().sprintf("/proc/%d/cwd", pid));
	
	i = readlink(qPrintable(cwdFile), buf, sizeof(buf));
	if(i >= 0) {
		buf[i] = '\0';
		ret = buf;
	}
	
	return ret;
}
